from __future__ import annotations

from typing import cast

import hashlib
import hmac

import google.protobuf.message

from ..ecc.curve import Curve
from ..ecc.djbec import CurvePublicKey
from ..exceptions import InvalidKeyException
from ..exceptions import InvalidMessageException
from ..identitykey import IdentityKey
from ..util.byteutil import ByteUtil
from . import omemo_pb2
from .ciphertextmessage import CiphertextMessage

MAC_LENGTH = 16


class OMEMOMessage(CiphertextMessage):
    def __init__(
        self,
        serialized: bytes,
        sender_ratchet_key: CurvePublicKey,
        counter: int,
        previous_counter: int,
        ciphertext: bytes,
    ):
        self._serialized = serialized
        self._sender_ratchet_key = sender_ratchet_key
        self._counter = counter
        self._previous_counter = previous_counter
        self._ciphertext = ciphertext

    @classmethod
    def new(
        cls,
        message_version: int,
        mac_key: bytes,
        ec_public_key_sender_ratchet_key: CurvePublicKey,
        counter: int,
        previous_counter: int,
        ciphertext: bytes,
        sender_identity_key: IdentityKey,
        receiver_identity_key: IdentityKey,
    ) -> OMEMOMessage:
        omemo_message = cast(
            OMEMOMessageProto, omemo_pb2.OMEMOMessage()  # pyright: ignore
        )

        omemo_message.dh_pub = ec_public_key_sender_ratchet_key.serialize()
        omemo_message.n = counter
        omemo_message.pn = previous_counter
        omemo_message.ciphertext = ciphertext
        omemo_message = omemo_message.SerializeToString()

        mac = cls.get_mac(
            sender_identity_key, receiver_identity_key, mac_key, omemo_message
        )

        authenticated_message = cast(
            OMEMOAuthenticatedMessageProto,
            omemo_pb2.OMEMOAuthenticatedMessage(),  # pyright: ignore
        )
        authenticated_message.mac = mac
        authenticated_message.message = omemo_message
        authenticated_message = authenticated_message.SerializeToString()

        return cls(
            authenticated_message,
            ec_public_key_sender_ratchet_key,
            counter,
            previous_counter,
            ciphertext,
        )

    @classmethod
    def from_bytes(cls, serialized: bytes) -> OMEMOMessage:
        authenticated_message = cast(
            OMEMOAuthenticatedMessageProto,
            omemo_pb2.OMEMOAuthenticatedMessage(),  # pyright: ignore
        )
        try:
            authenticated_message.ParseFromString(serialized)
        except google.protobuf.message.DecodeError as error:
            raise InvalidMessageException(str(error))

        omemo_message = cast(
            OMEMOMessageProto, omemo_pb2.OMEMOMessage()  # pyright: ignore
        )

        try:
            omemo_message.ParseFromString(authenticated_message.message)
        except google.protobuf.message.DecodeError as error:
            raise InvalidMessageException(str(error))

        try:
            sender_ratchet_key = Curve.decode_point(omemo_message.dh_pub)
            assert isinstance(sender_ratchet_key, CurvePublicKey)
        except InvalidKeyException as error:
            raise InvalidMessageException(str(error))

        return OMEMOMessage(
            serialized,
            sender_ratchet_key,
            omemo_message.n,
            omemo_message.pn,
            omemo_message.ciphertext,
        )

    def get_sender_ratchet_key(self) -> CurvePublicKey:
        return self._sender_ratchet_key

    def get_message_version(self) -> int:
        return 4

    def get_counter(self) -> int:
        return self._counter

    def get_body(self) -> bytes:
        return self._ciphertext

    def verify_mac(
        self,
        sender_identity_key: IdentityKey,
        receiver_identity_key: IdentityKey,
        mac_key: bytes,
    ):
        parts = ByteUtil.split(
            self._serialized,
            len(self._serialized) - MAC_LENGTH,
            MAC_LENGTH,
        )

        our_mac = self.get_mac(
            sender_identity_key, receiver_identity_key, mac_key, parts[0]
        )
        their_mac = parts[1]

        if our_mac != their_mac:
            raise InvalidMessageException("Bad Mac!")

    @classmethod
    def get_mac(
        cls,
        sender_identity_key: IdentityKey,
        receiver_identity_key: IdentityKey,
        mac_key: bytes,
        serialized: bytes,
    ) -> bytes:
        mac = hmac.new(mac_key, digestmod=hashlib.sha256)
        mac.update(sender_identity_key.get_public_key().serialize())
        mac.update(receiver_identity_key.get_public_key().serialize())
        mac.update(bytes(serialized))
        full_mac = mac.digest()
        return ByteUtil.trim(full_mac, MAC_LENGTH)

    def serialize(self) -> bytes:
        return self._serialized

    def get_type(self) -> int:
        return CiphertextMessage.WHISPER_TYPE


class OMEMOMessageProto(google.protobuf.message.Message):
    n: int
    pn: int
    dh_pub: bytes
    ciphertext: bytes


class OMEMOAuthenticatedMessageProto(google.protobuf.message.Message):
    mac: bytes
    message: bytes



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whisper.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rwhisper.proto\x12\ntextsecure\"b\n\x0eWhisperMessage\x12\x12\n\nratchetKey\x18\x01 \x01(\x0c\x12\x0f\n\x07\x63ounter\x18\x02 \x01(\r\x12\x17\n\x0fpreviousCounter\x18\x03 \x01(\r\x12\x12\n\nciphertext\x18\x04 \x01(\x0c\"\x8f\x01\n\x14PreKeyWhisperMessage\x12\x16\n\x0eregistrationId\x18\x05 \x01(\r\x12\x10\n\x08preKeyId\x18\x01 \x01(\r\x12\x16\n\x0esignedPreKeyId\x18\x06 \x01(\r\x12\x0f\n\x07\x62\x61seKey\x18\x02 \x01(\x0c\x12\x13\n\x0bidentityKey\x18\x03 \x01(\x0c\x12\x0f\n\x07message\x18\x04 \x01(\x0c\x42\x37\n&org.whispersystems.libaxolotl.protocolB\rWhisperProtos')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whisper_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n&org.whispersystems.libaxolotl.protocolB\rWhisperProtos'
  _WHISPERMESSAGE._serialized_start=29
  _WHISPERMESSAGE._serialized_end=127
  _PREKEYWHISPERMESSAGE._serialized_start=130
  _PREKEYWHISPERMESSAGE._serialized_end=273
# @@protoc_insertion_point(module_scope)



import abc


class CiphertextMessage(abc.ABC):
    SUPPORTED_VERSIONS = (3, 4)
    CURRENT_VERSION = 3

    WHISPER_TYPE = 2
    PREKEY_TYPE = 3
    SENDERKEY_TYPE = 4
    SENDERKEY_DISTRIBUTION_TYPE = 5

    # This should be the worst case (worse than V2).
    # So not always accurate, but good enough for padding.
    ENCRYPTED_MESSAGE_OVERHEAD = 53

    @abc.abstractmethod
    def serialize(self) -> bytes:
        return

    @abc.abstractmethod
    def get_type(self) -> int:
        return



from enum import IntEnum

NS_OMEMO_TMP = "eu.siacs.conversations.axolotl"
NS_OMEMO_2 = "urn:xmpp:omemo:2"

LEGACY_ENCODED_KEY_LENGTH = 33
ENCODED_KEY_LENGTH = 32
MAX_INT = 2**31 - 1


class OMEMOTrust(IntEnum):
    UNTRUSTED = 0
    VERIFIED = 1
    UNDECIDED = 2
    BLIND = 3



from __future__ import annotations

from typing import Optional

import hashlib
import hmac
import math


class HKDF:
    HASH_OUTPUT_SIZE = 32

    def __init__(self, session_version: int) -> None:
        if session_version not in (3, 4):
            raise AssertionError("Unknown version: %s " % session_version)

        self._session_version = session_version

    def get_session_version(self) -> int:
        return self._session_version

    def derive_secrets(
        self,
        input_key_material: bytes,
        info: bytes,
        output_length: int,
        salt: Optional[bytes] = None,
    ) -> bytes:
        salt = salt or bytes(self.HASH_OUTPUT_SIZE)
        prk = self._extract(salt, input_key_material)
        return self._expand(prk, info, output_length)

    def _extract(self, salt: bytes, input_key_material: bytes) -> bytes:
        mac = hmac.new(bytes(salt), digestmod=hashlib.sha256)
        mac.update(bytes(input_key_material))
        return mac.digest()

    def _expand(self, prk: bytes, info: bytes, output_size: int) -> bytes:
        iterations = int(math.ceil(float(output_size) / float(self.HASH_OUTPUT_SIZE)))
        mixin = b""
        results = bytearray()
        remaining_bytes = output_size

        for i in range(1, iterations + 1):
            mac = hmac.new(prk, digestmod=hashlib.sha256)
            mac.update(mixin)
            mac.update(info)
            updateChr = chr(i % 256)
            mac.update(updateChr.encode())

            stepResult = mac.digest()
            stepSize = min(remaining_bytes, len(stepResult))
            results.extend(stepResult[:stepSize])
            mixin = stepResult
            remaining_bytes -= stepSize

        return bytes(results)



from typing import Any

import logging
from collections.abc import MutableMapping


class SessionContextAdapter(logging.LoggerAdapter):  # pyright: ignore
    def process(
        self, msg: str, kwargs: MutableMapping[str, Any]
    ) -> tuple[str, MutableMapping[str, Any]]:
        assert self.extra is not None
        if self.extra["context"] is None:
            return msg, kwargs
        return f'({self.extra["context"]}) {msg}', kwargs



from __future__ import annotations

from typing import cast

import google.protobuf.message

from ..ecc.curve import Curve
from ..ecc.eckeypair import ECKeyPair
from . import storage_pb2


class PreKeyRecord:
    def __init__(self, structure: PreKeyRecordStructureProto) -> None:
        self._structure = structure

    @classmethod
    def new(
        cls,
        _id: int,
        ec_key_pair: ECKeyPair,
    ) -> PreKeyRecord:
        structure = cast(
            PreKeyRecordStructureProto,
            storage_pb2.PreKeyRecordStructure(),  # pyright: ignore
        )
        structure.id = _id
        structure.publicKey = ec_key_pair.get_public_key().serialize()
        structure.privateKey = ec_key_pair.get_private_key().serialize()
        return cls(structure)

    @classmethod
    def from_bytes(cls, serialized: bytes) -> PreKeyRecord:
        record = cast(
            PreKeyRecordStructureProto,
            storage_pb2.PreKeyRecordStructure(),  # pyright: ignore
        )
        record.ParseFromString(serialized)
        return cls(record)

    def get_id(self) -> int:
        return self._structure.id

    def get_key_pair(self):
        public_key = Curve.decode_point(self._structure.publicKey)
        private_key = Curve.decode_private_point(self._structure.privateKey)
        return ECKeyPair(public_key, private_key)

    def serialize(self) -> bytes:
        return self._structure.SerializeToString()


class PreKeyRecordStructureProto(google.protobuf.message.Message):
    id: int
    publicKey: bytes
    privateKey: bytes



from __future__ import annotations

from ..ecc.curve import Curve
from ..ecc.djbec import CurvePublicKey
from ..ecc.eckeypair import ECKeyPair
from ..kdf.derivedrootsecrets import DerivedRootSecrets
from ..kdf.hkdf import HKDF
from .chainkey import ChainKey


class RootKey:
    def __init__(self, kdf: HKDF, key: bytes) -> None:
        self._kdf = kdf
        self._key = key

    def get_key_bytes(self) -> bytes:
        return self._key

    def create_chain(
        self,
        ec_public_key_their_ratchet_key: CurvePublicKey,
        ec_key_pair_our_ratchet_key: ECKeyPair,
    ) -> tuple[RootKey, ChainKey]:
        if self._kdf.get_session_version() <= 3:
            domain_separator = "WhisperRatchet"
        else:
            domain_separator = "OMEMO Root Chain"

        shared_secret = Curve.calculate_agreement(
            ec_public_key_their_ratchet_key,
            ec_key_pair_our_ratchet_key.get_private_key(),
        )

        derived_secret_bytes = self._kdf.derive_secrets(
            shared_secret,
            domain_separator.encode(),
            DerivedRootSecrets.SIZE,
            salt=self._key,
        )

        derived_secrets = DerivedRootSecrets(derived_secret_bytes)
        new_root_key = RootKey(self._kdf, derived_secrets.get_root_key())
        new_chain_key = ChainKey(self._kdf, derived_secrets.get_chain_key(), 0)
        return (new_root_key, new_chain_key)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: storage.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rstorage.proto\x12\ntextsecure\"\xd3\x08\n\x10SessionStructure\x12\x16\n\x0esessionVersion\x18\x01 \x01(\r\x12\x1b\n\x13localIdentityPublic\x18\x02 \x01(\x0c\x12\x1c\n\x14remoteIdentityPublic\x18\x03 \x01(\x0c\x12\x0f\n\x07rootKey\x18\x04 \x01(\x0c\x12\x17\n\x0fpreviousCounter\x18\x05 \x01(\r\x12\x37\n\x0bsenderChain\x18\x06 \x01(\x0b\x32\".textsecure.SessionStructure.Chain\x12:\n\x0ereceiverChains\x18\x07 \x03(\x0b\x32\".textsecure.SessionStructure.Chain\x12K\n\x12pendingKeyExchange\x18\x08 \x01(\x0b\x32/.textsecure.SessionStructure.PendingKeyExchange\x12\x41\n\rpendingPreKey\x18\t \x01(\x0b\x32*.textsecure.SessionStructure.PendingPreKey\x12\x1c\n\x14remoteRegistrationId\x18\n \x01(\r\x12\x1b\n\x13localRegistrationId\x18\x0b \x01(\r\x12\x14\n\x0cneedsRefresh\x18\x0c \x01(\x08\x12\x14\n\x0c\x61liceBaseKey\x18\r \x01(\x0c\x1a\xb9\x02\n\x05\x43hain\x12\x18\n\x10senderRatchetKey\x18\x01 \x01(\x0c\x12\x1f\n\x17senderRatchetKeyPrivate\x18\x02 \x01(\x0c\x12=\n\x08\x63hainKey\x18\x03 \x01(\x0b\x32+.textsecure.SessionStructure.Chain.ChainKey\x12\x42\n\x0bmessageKeys\x18\x04 \x03(\x0b\x32-.textsecure.SessionStructure.Chain.MessageKey\x1a&\n\x08\x43hainKey\x12\r\n\x05index\x18\x01 \x01(\r\x12\x0b\n\x03key\x18\x02 \x01(\x0c\x1aJ\n\nMessageKey\x12\r\n\x05index\x18\x01 \x01(\r\x12\x11\n\tcipherKey\x18\x02 \x01(\x0c\x12\x0e\n\x06macKey\x18\x03 \x01(\x0c\x12\n\n\x02iv\x18\x04 \x01(\x0c\x1a\xcd\x01\n\x12PendingKeyExchange\x12\x10\n\x08sequence\x18\x01 \x01(\r\x12\x14\n\x0clocalBaseKey\x18\x02 \x01(\x0c\x12\x1b\n\x13localBaseKeyPrivate\x18\x03 \x01(\x0c\x12\x17\n\x0flocalRatchetKey\x18\x04 \x01(\x0c\x12\x1e\n\x16localRatchetKeyPrivate\x18\x05 \x01(\x0c\x12\x18\n\x10localIdentityKey\x18\x07 \x01(\x0c\x12\x1f\n\x17localIdentityKeyPrivate\x18\x08 \x01(\x0c\x1aJ\n\rPendingPreKey\x12\x10\n\x08preKeyId\x18\x01 \x01(\r\x12\x16\n\x0esignedPreKeyId\x18\x03 \x01(\x05\x12\x0f\n\x07\x62\x61seKey\x18\x02 \x01(\x0c\"\x7f\n\x0fRecordStructure\x12\x34\n\x0e\x63urrentSession\x18\x01 \x01(\x0b\x32\x1c.textsecure.SessionStructure\x12\x36\n\x10previousSessions\x18\x02 \x03(\x0b\x32\x1c.textsecure.SessionStructure\"J\n\x15PreKeyRecordStructure\x12\n\n\x02id\x18\x01 \x01(\r\x12\x11\n\tpublicKey\x18\x02 \x01(\x0c\x12\x12\n\nprivateKey\x18\x03 \x01(\x0c\"v\n\x1bSignedPreKeyRecordStructure\x12\n\n\x02id\x18\x01 \x01(\r\x12\x11\n\tpublicKey\x18\x02 \x01(\x0c\x12\x12\n\nprivateKey\x18\x03 \x01(\x0c\x12\x11\n\tsignature\x18\x04 \x01(\x0c\x12\x11\n\ttimestamp\x18\x05 \x01(\x06\"A\n\x18IdentityKeyPairStructure\x12\x11\n\tpublicKey\x18\x01 \x01(\x0c\x12\x12\n\nprivateKey\x18\x02 \x01(\x0c\"\xb8\x03\n\x17SenderKeyStateStructure\x12\x13\n\x0bsenderKeyId\x18\x01 \x01(\r\x12J\n\x0esenderChainKey\x18\x02 \x01(\x0b\x32\x32.textsecure.SenderKeyStateStructure.SenderChainKey\x12N\n\x10senderSigningKey\x18\x03 \x01(\x0b\x32\x34.textsecure.SenderKeyStateStructure.SenderSigningKey\x12O\n\x11senderMessageKeys\x18\x04 \x03(\x0b\x32\x34.textsecure.SenderKeyStateStructure.SenderMessageKey\x1a\x31\n\x0eSenderChainKey\x12\x11\n\titeration\x18\x01 \x01(\r\x12\x0c\n\x04seed\x18\x02 \x01(\x0c\x1a\x33\n\x10SenderMessageKey\x12\x11\n\titeration\x18\x01 \x01(\r\x12\x0c\n\x04seed\x18\x02 \x01(\x0c\x1a\x33\n\x10SenderSigningKey\x12\x0e\n\x06public\x18\x01 \x01(\x0c\x12\x0f\n\x07private\x18\x02 \x01(\x0c\"X\n\x18SenderKeyRecordStructure\x12<\n\x0fsenderKeyStates\x18\x01 \x03(\x0b\x32#.textsecure.SenderKeyStateStructureB4\n#org.whispersystems.libaxolotl.stateB\rStorageProtos')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'storage_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n#org.whispersystems.libaxolotl.stateB\rStorageProtos'
  _SESSIONSTRUCTURE._serialized_start=30
  _SESSIONSTRUCTURE._serialized_end=1137
  _SESSIONSTRUCTURE_CHAIN._serialized_start=540
  _SESSIONSTRUCTURE_CHAIN._serialized_end=853
  _SESSIONSTRUCTURE_CHAIN_CHAINKEY._serialized_start=739
  _SESSIONSTRUCTURE_CHAIN_CHAINKEY._serialized_end=777
  _SESSIONSTRUCTURE_CHAIN_MESSAGEKEY._serialized_start=779
  _SESSIONSTRUCTURE_CHAIN_MESSAGEKEY._serialized_end=853
  _SESSIONSTRUCTURE_PENDINGKEYEXCHANGE._serialized_start=856
  _SESSIONSTRUCTURE_PENDINGKEYEXCHANGE._serialized_end=1061
  _SESSIONSTRUCTURE_PENDINGPREKEY._serialized_start=1063
  _SESSIONSTRUCTURE_PENDINGPREKEY._serialized_end=1137
  _RECORDSTRUCTURE._serialized_start=1139
  _RECORDSTRUCTURE._serialized_end=1266
  _PREKEYRECORDSTRUCTURE._serialized_start=1268
  _PREKEYRECORDSTRUCTURE._serialized_end=1342
  _SIGNEDPREKEYRECORDSTRUCTURE._serialized_start=1344
  _SIGNEDPREKEYRECORDSTRUCTURE._serialized_end=1462
  _IDENTITYKEYPAIRSTRUCTURE._serialized_start=1464
  _IDENTITYKEYPAIRSTRUCTURE._serialized_end=1529
  _SENDERKEYSTATESTRUCTURE._serialized_start=1532
  _SENDERKEYSTATESTRUCTURE._serialized_end=1972
  _SENDERKEYSTATESTRUCTURE_SENDERCHAINKEY._serialized_start=1817
  _SENDERKEYSTATESTRUCTURE_SENDERCHAINKEY._serialized_end=1866
  _SENDERKEYSTATESTRUCTURE_SENDERMESSAGEKEY._serialized_start=1868
  _SENDERKEYSTATESTRUCTURE_SENDERMESSAGEKEY._serialized_end=1919
  _SENDERKEYSTATESTRUCTURE_SENDERSIGNINGKEY._serialized_start=1921
  _SENDERKEYSTATESTRUCTURE_SENDERSIGNINGKEY._serialized_end=1972
  _SENDERKEYRECORDSTRUCTURE._serialized_start=1974
  _SENDERKEYRECORDSTRUCTURE._serialized_end=2062
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import Union

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers import modes

from .ecc.curve import Curve
from .ecc.djbec import CurvePublicKey
from .exceptions import DuplicateMessageException
from .exceptions import InvalidMessageException
from .exceptions import NoSessionException
from .kdf.messagekeys import MessageKeys
from .protocol.prekeywhispermessage import PreKeyWhisperMessage
from .protocol.whispermessage import WhisperMessage
from .ratchet.chainkey import ChainKey
from .sessionbuilder import SessionBuilder
from .state.sessionrecord import SessionRecord
from .state.sessionstate import SessionState
from .state.store import Store


class SessionCipher:
    def __init__(self, store: Store, recipient_id: str, device_id: int) -> None:
        self._store = store
        self._recipient_id = recipient_id
        self._device_id = device_id
        self._session_builder = SessionBuilder(
            store,
            recipient_id,
            device_id,
        )

    def encrypt(
        self, padded_message: Union[bytes, str]
    ) -> Union[WhisperMessage, PreKeyWhisperMessage]:
        if isinstance(padded_message, str):
            padded_message = padded_message.encode()

        session_record = self._store.load_session(self._recipient_id, self._device_id)
        session_state = session_record.get_session_state()
        chain_key = session_state.get_sender_chain_key()
        message_keys = chain_key.get_message_keys()
        sender_ephemeral = session_state.get_sender_ratchet_key()
        previous_counter = session_state.get_previous_counter()
        session_version = session_state.get_session_version()

        ciphertext_body = self.get_ciphertext(
            session_version, message_keys, padded_message
        )

        ciphertext_message = WhisperMessage.new(
            session_version,
            message_keys.get_mac_key(),
            sender_ephemeral,
            chain_key.get_index(),
            previous_counter,
            ciphertext_body,
            session_state.get_local_identity_key(),
            session_state.get_remote_identity_key(),
        )

        if session_state.has_unacknowledged_pre_key_message():
            items = session_state.get_unacknowledged_pre_key_message_items()
            our_device_id = session_state.get_our_device_id()

            ciphertext_message = PreKeyWhisperMessage.new(
                session_version,
                our_device_id,
                items.get_pre_key_id(),
                items.get_signed_pre_key_id(),
                items.get_base_key(),
                session_state.get_local_identity_key(),
                ciphertext_message,
            )

        session_state.set_sender_chain_key(chain_key.get_next_chain_key())
        self._store.store_session(self._recipient_id, self._device_id, session_record)

        return ciphertext_message

    def decrypt_msg(self, ciphertext: WhisperMessage) -> bytes:
        if not self._store.contains_session(self._recipient_id, self._device_id):
            raise NoSessionException(
                "No session for: %s, %s" % (self._recipient_id, self._device_id)
            )

        session_record = self._store.load_session(self._recipient_id, self._device_id)
        plaintext = self.decrypt_with_session_record(session_record, ciphertext)

        self._store.store_session(self._recipient_id, self._device_id, session_record)

        return plaintext

    def decrypt_pkmsg(self, ciphertext: PreKeyWhisperMessage) -> bytes:
        session_record = self._store.load_session(self._recipient_id, self._device_id)
        unsigned_pre_key_id = self._session_builder.process(session_record, ciphertext)
        plaintext = self.decrypt_with_session_record(
            session_record, ciphertext.get_whisper_message()
        )

        self._store.store_session(self._recipient_id, self._device_id, session_record)

        if unsigned_pre_key_id is not None:
            self._store.remove_pre_key(unsigned_pre_key_id)

        return plaintext

    def decrypt_with_session_record(
        self, session_record: SessionRecord, cipher_text: WhisperMessage
    ) -> bytes:
        previous_states = session_record.get_previous_session_states()
        exceptions: list[Exception] = []
        try:
            session_state = SessionState(session_record.get_session_state())
            plaintext = self.decrypt_with_session_state(session_state, cipher_text)
            session_record.set_state(session_state)
            return plaintext
        except InvalidMessageException as e:
            exceptions.append(e)

        for i in range(0, len(previous_states)):
            previous_state = previous_states[i]
            try:
                promotedState = SessionState(previous_state)
                plaintext = self.decrypt_with_session_state(promotedState, cipher_text)
                previous_states.pop(i)
                session_record.promote_state(promotedState)
                return plaintext
            except InvalidMessageException as e:
                exceptions.append(e)

        raise InvalidMessageException("No valid sessions", exceptions)

    def decrypt_with_session_state(
        self, session_state: SessionState, ciphertext_message: WhisperMessage
    ) -> bytes:
        if not session_state.has_sender_chain():
            raise InvalidMessageException("Uninitialized session!")

        message_version = ciphertext_message.get_message_version()
        if message_version != session_state.get_session_version():
            raise InvalidMessageException(
                "Message version %s, but session version %s"
                % (
                    ciphertext_message.get_message_version,
                    session_state.get_session_version(),
                )
            )

        their_ephemeral = ciphertext_message.get_sender_ratchet_key()
        counter = ciphertext_message.get_counter()
        chain_key = self.get_or_create_chain_key(session_state, their_ephemeral)
        message_keys = self.get_or_create_message_keys(
            session_state, their_ephemeral, chain_key, counter
        )

        ciphertext_message.verify_mac(
            session_state.get_remote_identity_key(),
            session_state.get_local_identity_key(),
            message_keys.get_mac_key(),
        )

        plaintext = self.get_plaintext(
            message_version, message_keys, ciphertext_message.get_body()
        )
        session_state.clear_unacknowledged_pre_key_message()

        return plaintext

    def get_or_create_chain_key(
        self,
        session_state: SessionState,
        ec_publick_key_their_ephemeral: CurvePublicKey,
    ) -> ChainKey:
        their_ephemeral = ec_publick_key_their_ephemeral
        if session_state.has_receiver_chain(their_ephemeral):
            return session_state.get_receiver_chain_key(their_ephemeral)
        else:
            root_key = session_state.get_root_key()
            our_ephemeral = session_state.get_sender_ratchet_key_pair()
            receiver_chain = root_key.create_chain(their_ephemeral, our_ephemeral)
            our_new_ephemeral = Curve.generate_key_pair()
            sender_chain = receiver_chain[0].create_chain(
                their_ephemeral, our_new_ephemeral
            )

            session_state.set_root_key(sender_chain[0])
            session_state.add_receiver_chain(their_ephemeral, receiver_chain[1])
            session_state.set_previous_counter(
                max(session_state.get_sender_chain_key().get_index() - 1, 0)
            )
            session_state.set_sender_chain(our_new_ephemeral, sender_chain[1])
            return receiver_chain[1]

    def get_or_create_message_keys(
        self,
        session_state: SessionState,
        ec_public_key_their_ephemeral: CurvePublicKey,
        chain_key: ChainKey,
        counter: int,
    ) -> MessageKeys:
        their_ephemeral = ec_public_key_their_ephemeral
        if chain_key.get_index() > counter:
            if session_state.has_message_keys(their_ephemeral, counter):
                return session_state.remove_message_keys(their_ephemeral, counter)
            else:
                raise DuplicateMessageException(
                    "Received message with old counter: %s, %s"
                    % (chain_key.get_index(), counter)
                )

        if counter - chain_key.get_index() > 2000:
            raise InvalidMessageException("Over 2000 messages into the future!")

        while chain_key.get_index() < counter:
            message_keys = chain_key.get_message_keys()
            session_state.set_message_keys(their_ephemeral, message_keys)
            chain_key = chain_key.get_next_chain_key()

        session_state.set_receiver_chain_key(
            their_ephemeral, chain_key.get_next_chain_key()
        )
        return chain_key.get_message_keys()

    def get_ciphertext(
        self, version: int, message_keys: MessageKeys, plain_text: bytes
    ) -> bytes:
        cipher = AESCipher(message_keys.get_cipher_key(), message_keys.get_iv())
        return cipher.encrypt(plain_text)

    def get_plaintext(
        self, version: int, message_keys: MessageKeys, cipher_text: bytes
    ) -> bytes:
        cipher = AESCipher(message_keys.get_cipher_key(), message_keys.get_iv())
        return cipher.decrypt(cipher_text)


class AESCipher:
    def __init__(self, key: bytes, iv: bytes) -> None:
        self._key = key
        self._iv = iv
        self._cipher = Cipher(
            algorithms.AES(key), modes.CBC(iv), backend=default_backend()
        )

    def unpad(self, data: bytes) -> bytes:
        unpad_length = data[-1]
        if isinstance(unpad_length, int):  # pyright: ignore
            cmp = bytes([data[-unpad_length]] * unpad_length)
        else:
            raise AssertionError("unpad_length is not integer")
            # unpad_length = ord(unpad_length)
            # cmp = data[-unpad_length] * unpad_length
        if data[-unpad_length:] != cmp:
            raise ValueError("Data not properly padded \n %s" % data)

        return data[0:-unpad_length]

    def pad(self, s: bytes) -> bytes:
        return s + ((16 - len(s) % 16) * chr(16 - len(s) % 16)).encode()

    def encrypt(self, raw: bytes) -> bytes:
        raw_padded = self.pad(raw)
        encryptor = self._cipher.encryptor()
        try:
            return encryptor.update(raw_padded) + encryptor.finalize()
        except ValueError:
            raise

    def decrypt(self, enc: bytes) -> bytes:
        decryptor = self._cipher.decryptor()
        return self.unpad(decryptor.update(enc) + decryptor.finalize())



from __future__ import annotations

from typing import Any

import binascii

from .. import curve
from ..const import ENCODED_KEY_LENGTH
from ..const import LEGACY_ENCODED_KEY_LENGTH
from ..exceptions import InvalidKeyException
from ..util.byteutil import ByteUtil
from .ec import ECPrivateKey
from .ec import ECPublicKey

DJB_TYPE = 0x05


class DjbECPublicKey(ECPublicKey):
    def __init__(self, _bytes: bytes) -> None:
        self._public_key = _bytes

    def get_bytes(self) -> bytes:
        return self._public_key

    def get_public_key(self) -> bytes:
        return self._public_key

    def __eq__(self, other: Any) -> bool:
        return self._public_key == other.get_public_key()

    def __lt__(self, other: Any) -> bool:
        my_val = int(binascii.hexlify(self._public_key), 16)
        other_val = int(binascii.hexlify(other.get_public_key()), 16)

        return my_val < other_val

    def __cmp__(self, other: Any) -> int:
        my_val = int(binascii.hexlify(self._public_key), 16)
        other_val = int(binascii.hexlify(other.get_public_key()), 16)

        if my_val < other_val:
            return -1
        elif my_val == other_val:
            return 0
        else:
            return 1


class CurvePublicKey(DjbECPublicKey):
    @classmethod
    def from_bytes(cls, bytes_: bytes) -> CurvePublicKey:
        key_length = len(bytes_)
        if key_length == LEGACY_ENCODED_KEY_LENGTH and bytes_[0] == 0x05:
            return cls(bytes_[1:])

        if key_length == ENCODED_KEY_LENGTH:
            return cls(bytes_)

        raise InvalidKeyException("Unknown key type or length: %s" % bytes_)

    def serialize(self) -> bytes:
        return ByteUtil.combine([DJB_TYPE], self._public_key)

    def to_ed(self) -> EdPublicKey:
        return EdPublicKey(curve.convert_curve_to_ed_pubkey(self._public_key))


class EdPublicKey(DjbECPublicKey):
    @classmethod
    def from_bytes(cls, bytes_: bytes) -> EdPublicKey:
        key_length = len(bytes_)
        if key_length == ENCODED_KEY_LENGTH:
            return cls(bytes_)

        raise InvalidKeyException("Unknown key type or length: %s" % bytes_)

    def to_curve(self) -> CurvePublicKey:
        return CurvePublicKey(curve.convert_ed_to_curve_pubkey(self._public_key))

    def serialize(self) -> bytes:
        return self._public_key


class DjbECPrivateKey(ECPrivateKey):
    def __init__(self, private_key: bytes) -> None:
        self._private_key = private_key

    def get_private_key(self) -> bytes:
        return self._private_key

    def serialize(self) -> bytes:
        return self._private_key

    def __eq__(self, other: Any) -> bool:
        return self._private_key == other.get_private_key()



from __future__ import annotations

from typing import Union

import binascii
import textwrap

from .ecc.curve import Curve
from .ecc.djbec import CurvePublicKey


class IdentityKey:
    def __init__(
        self,
        ec_pub_key_or_bytes: Union[CurvePublicKey, bytes],
    ) -> None:
        if isinstance(ec_pub_key_or_bytes, CurvePublicKey):
            self._public_key = ec_pub_key_or_bytes
        else:
            self._public_key = Curve.decode_point(ec_pub_key_or_bytes)

    def get_public_key(self) -> CurvePublicKey:
        return self._public_key

    def serialize(self) -> bytes:
        return self._public_key.serialize()

    def get_fingerprint(self, formatted: bool = False) -> str:
        public_key = self.serialize()
        fingerprint = binascii.hexlify(public_key).decode()[2:]
        if not formatted:
            return fingerprint
        fplen = len(fingerprint)
        wordsize = fplen // 8
        buf = ""
        for word in range(0, fplen, wordsize):
            buf += f"{fingerprint[word : word + wordsize]} "
        buf = textwrap.fill(buf, width=36)
        return buf.rstrip().upper()

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, IdentityKey)
        return self._public_key == other.get_public_key()

    def __hash__(self) -> int:
        return hash(self._public_key.serialize())



from ..util.byteutil import ByteUtil


class DerivedMessageSecrets:
    SIZE = 80
    CIPHER_KEY_LENGTH = 32
    MAC_KEY_LENGTH = 32
    IV_LENGTH = 16

    def __init__(self, okm: bytes) -> None:
        keys = ByteUtil.split(
            okm, self.CIPHER_KEY_LENGTH, self.MAC_KEY_LENGTH, self.IV_LENGTH
        )
        self._cipher_key = keys[0]  # AES
        self._mac_key = keys[1]  # sha256
        self._iv = keys[2]

    def get_cipher_key(self) -> bytes:
        return self._cipher_key

    def get_mac_key(self) -> bytes:
        return self._mac_key

    def get_iv(self) -> bytes:
        return self._iv



from __future__ import annotations

from typing import Any

import logging
from collections import defaultdict
from collections.abc import Callable


class Observable:
    def __init__(self) -> None:
        self._log = logging.getLogger(__name__)
        self._callbacks: defaultdict[str, list[Callable[..., Any]]] = defaultdict(list)

    def register_signal(self, signal_name: str, func: Callable[..., Any]) -> None:
        self._callbacks[signal_name].append(func)

    def _unregister_signals(self) -> None:
        self._callbacks = defaultdict(list)

    def _notify(self, signal_name: str, *args: Any, **kwargs: dict[str, Any]) -> None:
        self._log.info("Signal: %s", signal_name)
        callbacks = self._callbacks.get(signal_name, [])
        for func in callbacks:
            func(self, signal_name, *args, **kwargs)



from __future__ import annotations

from ..const import MAX_INT
from ..const import NS_OMEMO_2
from ..const import NS_OMEMO_TMP
from ..ecc.djbec import CurvePublicKey
from ..ecc.djbec import EdPublicKey
from ..exceptions import BundleValidationError
from ..identitykey import IdentityKey
from ..structs import OMEMOBundleProto


class PreKeyBundle:
    def __init__(
        self,
        remote_device_id: int,
        namespace: str,
        pre_key_id: int,
        ec_public_key_pre_key_public: CurvePublicKey,
        signed_pre_key_id: int,
        ec_public_key_signed_pre_key_public: CurvePublicKey,
        signed_pre_key_signature: bytes,
        identity_key: IdentityKey,
    ) -> None:
        self._remote_device_id = remote_device_id
        self._namespace = namespace
        self._pre_key_id = pre_key_id
        self._pre_key_public = ec_public_key_pre_key_public
        self._signed_pre_key_id = signed_pre_key_id
        self._signed_pre_key_public = ec_public_key_signed_pre_key_public
        self._signed_pre_key_signature = signed_pre_key_signature
        self._identity_key = identity_key

    @classmethod
    def from_proto(cls, bundle: OMEMOBundleProto) -> PreKeyBundle:
        prekey = bundle.pick_prekey()
        otpk = CurvePublicKey.from_bytes(prekey["key"])
        spk = CurvePublicKey.from_bytes(bundle.spk["key"])

        ns = bundle.namespace
        if ns == NS_OMEMO_TMP:
            ik_pub = CurvePublicKey.from_bytes(bundle.ik)
        elif ns == NS_OMEMO_2:
            ik_pub = EdPublicKey.from_bytes(bundle.ik).to_curve()
        else:
            raise BundleValidationError("Unknown namespace on bundle: %s", ns)

        ik = IdentityKey(ik_pub)

        if not 1 <= bundle.device_id <= MAX_INT:
            raise BundleValidationError("Device id out of range")

        if not 1 <= prekey["id"] <= MAX_INT:
            raise BundleValidationError("Prekey id out of range")

        if not 0 <= bundle.spk["id"] <= MAX_INT:
            # Allow 0 to stay backwards compatible
            raise BundleValidationError("Signed pre key id out of range")

        return cls(
            bundle.device_id,
            bundle.namespace,
            prekey["id"],
            otpk,
            bundle.spk["id"],
            spk,
            bundle.spk_signature,
            ik,
        )

    def get_remote_device_id(self) -> int:
        return self._remote_device_id

    def get_namespace(self) -> str:
        return self._namespace

    def get_pre_key_id(self) -> int:
        return self._pre_key_id

    def get_pre_key(self) -> CurvePublicKey:
        return self._pre_key_public

    def get_signed_pre_key_id(self) -> int:
        return self._signed_pre_key_id

    def get_signed_pre_key(self) -> CurvePublicKey:
        return self._signed_pre_key_public

    def get_signed_pre_key_signature(self) -> bytes:
        return self._signed_pre_key_signature

    def get_identity_key(self) -> IdentityKey:
        return self._identity_key

    def get_session_version(self) -> int:
        if self._namespace == NS_OMEMO_TMP:
            return 3

        elif self._namespace == NS_OMEMO_2:
            return 4

        else:
            raise AssertionError("Unknown session version")



from __future__ import annotations

from typing import Optional

import logging

from .ecc.curve import Curve
from .exceptions import InvalidKeyException
from .exceptions import UntrustedIdentityException
from .protocol.prekeywhispermessage import PreKeyWhisperMessage
from .ratchet.aliceparameters import AliceParameters
from .ratchet.bobparameters import BobParameters
from .ratchet.ratchetingsession import RatchetingSession
from .state.prekeybundle import PreKeyBundle
from .state.sessionrecord import SessionRecord
from .state.store import Store

log = logging.getLogger(__name__)


class SessionBuilder:
    def __init__(
        self,
        store: Store,
        recipient_id: str,
        device_id: int,
    ) -> None:
        self._store = store
        self._recipient_id = recipient_id
        self._device_id = device_id

    def process(
        self, session_record: SessionRecord, message: PreKeyWhisperMessage
    ) -> Optional[int]:
        message_version = message.get_message_version()
        their_identity_key = message.get_identity_key()

        unsigned_pre_key_id = None

        if not self._store.is_trusted_identity(self._recipient_id, their_identity_key):
            raise UntrustedIdentityException(self._recipient_id, their_identity_key)

        if message_version in (3, 4):
            unsigned_pre_key_id = self.process_message(session_record, message)
        else:
            raise AssertionError("Unknown version %s" % message_version)

        self._store.save_identity(self._recipient_id, their_identity_key)

        return unsigned_pre_key_id

    def process_message(
        self, session_record: SessionRecord, message: PreKeyWhisperMessage
    ) -> Optional[int]:
        if session_record.has_session_state(
            message.get_message_version(), message.get_base_key().serialize()
        ):
            log.warning(
                "We've already setupgetMessageVersion a "
                "session for this V3 message, letting bundled "
                "message fall through..."
            )
            return None

        our_signed_pre_key = self._store.load_signed_pre_key(
            message.get_signed_pre_key_id()
        )
        our_signed_pre_key_pair = our_signed_pre_key.get_key_pair()
        our_one_time_prekey = self._store.load_pre_key(
            message.get_pre_key_id()
        ).get_key_pair()

        parameters = BobParameters(
            self._store.get_identity_key_pair(),
            our_signed_pre_key_pair,
            our_signed_pre_key_pair,
            our_one_time_prekey,
            message.get_identity_key(),
            message.get_base_key(),
        )

        if not session_record.is_fresh():
            session_record.archive_current_state()

        RatchetingSession.initialize_session_as_bob(
            session_record.get_session_state(),
            message.get_message_version(),
            parameters,
        )
        session_record.get_session_state().set_our_device_id(
            self._store.get_our_device_id()
        )
        session_record.get_session_state().set_remote_device_id(message.get_device_id())
        session_record.get_session_state().set_alice_base_key(
            message.get_base_key().serialize()
        )

        return message.get_pre_key_id()

    def process_pre_key_bundle(self, bundle: PreKeyBundle) -> None:
        if not self._store.is_trusted_identity(
            self._recipient_id, bundle.get_identity_key()
        ):
            raise UntrustedIdentityException(
                self._recipient_id, bundle.get_identity_key()
            )

        if not Curve.verify_signature(
            bundle.get_identity_key().get_public_key(),
            bundle.get_signed_pre_key().serialize(),
            bundle.get_signed_pre_key_signature(),
        ):
            raise InvalidKeyException("Invalid signature on device key!")

        session_record = self._store.load_session(self._recipient_id, self._device_id)
        our_base_key = Curve.generate_key_pair()
        their_signed_pre_key = bundle.get_signed_pre_key()
        their_one_time_pre_key = bundle.get_pre_key()
        their_one_time_pre_key_id = bundle.get_pre_key_id()

        parameters = AliceParameters(
            self._store.get_identity_key_pair(),
            our_base_key,
            bundle.get_identity_key(),
            their_signed_pre_key,
            their_signed_pre_key,
            their_one_time_pre_key,
        )

        if not session_record.is_fresh():
            session_record.archive_current_state()

        RatchetingSession.initialize_session_as_alice(
            session_record.get_session_state(),
            bundle.get_session_version(),
            parameters,
        )

        session_record.get_session_state().set_unacknowledged_pre_key_message(
            their_one_time_pre_key_id,
            bundle.get_signed_pre_key_id(),
            our_base_key.get_public_key(),
        )
        session_record.get_session_state().set_our_device_id(
            self._store.get_our_device_id()
        )
        session_record.get_session_state().set_remote_device_id(
            bundle.get_remote_device_id()
        )
        session_record.get_session_state().set_alice_base_key(
            our_base_key.get_public_key().serialize()
        )
        self._store.store_session(self._recipient_id, self._device_id, session_record)
        self._store.save_identity(self._recipient_id, bundle.get_identity_key())
